
#include "GenericTypeDefs.h"
#include "Compiler.h"
#include "HardwareProfile.h"
#include "game.h"
#include "pixmaps.h"

#define SCR_WIDTH 16
#define SCR_HEIGHT 32

#define SCR_SIZE 64
extern BYTE SCREENMEM[SCR_SIZE]; 
BYTE OFFSCREEN[SCR_SIZE]; 

#define SCORE_START (SCR_HEIGHT - 5)

// 1. Consider the device in a portrait orientation
//    with the USB port at the top. 
// 2. Say (0, 0) is the top left corner.
// 
// When we set SCREENMEM[i] = (1 << j)
// we light up pixel (x, y) where:
//    x = 15 - (i / 4) 
//    y = 31 - (((i % 4) * 8) + (7 - j))
//
// Therefore, convert (x, y) to (i, j):
//    i = ((15 - x) * 4) + ((31 - y) / 8)
//    j = 7 - ((31 - y) % 8)

#define get_scr_i(x, y) ((15 - (x)) * 4) + (((y)) / 8)
#define get_scr_j(x, y) (7 - (((y)) % 8))
#define set_pixel(x, y) \
    OFFSCREEN[get_scr_i((x), (y))] |= (1 << get_scr_j((x), (y)))
#define clr_pixel(x, y) \
    OFFSCREEN[get_scr_i((x), (y))] &= ~(1 << get_scr_j((x), (y)))
#define get_pixel(x, y) \
    (OFFSCREEN[get_scr_i((x), (y))] & (1 << get_scr_j((x), (y))))

// This is the push button on the back
#define is_button_pressed() (((BYTE) PORTCbits.RC1) == 0)

// This is a button connected to RE0
#define is_tilted() (((BYTE) PORTEbits.RE0) == 0)

// The speed at which the interrupt counter increments
// Set in basic.asm as a divider of the master clock
#define CATCHUP_FREQUENCY       ((WORD) 735) // Hz

#define MIN_PERIOD              26      // increments
#define MAX_PERIOD              400     // increments
#define SCROLL_PERIOD           60      // increments
#define FLASH_TIME              (CATCHUP_FREQUENCY / 10)
#define HOLD_DOWN_TIME          (CATCHUP_FREQUENCY * 2)
#define SCREEN_TIME             (CATCHUP_FREQUENCY * 7)


void halt_and_catch_fire(BYTE info);
BYTE GetElapsedLines(void);

static DWORD drng_value;
static WORD timing_error = 0;
static WORD min_slack;
static WORD aux_timer;
static WORD game_number = 0;

static void wait_period(WORD period) 
{
    WORD time_used = 0;

    time_used += (WORD) GetElapsedLines();
    time_used += timing_error;

    if (time_used >= period) {
        min_slack = 0;
    } else {
        WORD s = period - time_used;
        if (s < min_slack) min_slack = s;

        while (time_used < period) {
            time_used += (WORD) GetElapsedLines();
        }
    }
    time_used -= period;
    timing_error = time_used;
}

static const DWORD digitable[12] = {
        0x75557L, // 0
        0x11111L, // 1
        0x71747L, // 2
        0x71717L, // 3
        0x44571L, // 4
        0x74717L, // 5
        0x74757L, // 6
        0x71111L, // 7
        0x75757L, // 8
        0x75711L, // 9
        0x44447L, // L
        0x74457L, // G
        };

static void draw_digit(BYTE x, BYTE y, BYTE digit)
{
    BYTE x1, y1;
    DWORD all_data = digitable[digit];
    BYTE ored = 0;

    all_data = digitable[digit];
    for (y1 = 0; y1 < 5; y1++) {
        BYTE line = ((BYTE) (all_data >> (DWORD) 16L)) & 0xf;
        for (x1 = 0; x1 < 3; x1++) {
            if ((line & 0x4) && ((x + x1) < SCR_WIDTH)) {
                set_pixel(x + x1, y + y1);
            }
            line = line << 1;
        }
        all_data = all_data << (DWORD) 4L;
    }
}

static BYTE draw_number(BYTE x, BYTE y, WORD number)
{
    WORD here = (number % (WORD) 10L);
    number /= (WORD) 10L;

    if (number) {
        x = draw_number(x, y, number);
    }
    draw_digit(x, y, here);
    return x + 4;
}

static void draw_blank(void)
{
    BYTE i;
    for (i = 0; i < SCR_SIZE; i++) {
        OFFSCREEN[i] = 0;
    }
}

static BYTE drng(void)
{
    drng_value = ((((DWORD) drng_value & (DWORD) 0xffffL) * (DWORD) 18000L) 
            + ((DWORD) drng_value >> (DWORD) 16L));
    return (BYTE) drng_value;
}

static void flip(void)
{
    BYTE i;

    for (i = 0; i < SCR_SIZE; i++) {
        SCREENMEM[i] = OFFSCREEN[i];
    }
}

static void scroll_in(void)
{
    BYTE i, x, y;
    BYTE go_up = (drng() & 1);

    for (i = 0; i < SCR_HEIGHT; i++) {
        for (x = 0; x < SCR_WIDTH; x++) {
            BYTE q = x * 4;

            if (go_up) {
                for (y = 3; y > 0; y--) {
                    SCREENMEM[y + q] = SCREENMEM[y + q] >> 1;
                    SCREENMEM[y + q] |= SCREENMEM[y - 1 + q] << 7;
                }
                SCREENMEM[q] = SCREENMEM[q] >> 1;
                SCREENMEM[q] |= OFFSCREEN[3 + q] << 7;
                for (y = 3; y > 0; y--) {
                    OFFSCREEN[y + q] = OFFSCREEN[y + q] >> 1;
                    OFFSCREEN[y + q] |= OFFSCREEN[y - 1 + q] << 7;
                }
                OFFSCREEN[q] = OFFSCREEN[q] >> 1;
            } else {
                for (y = 0; y < 3; y++) {
                    SCREENMEM[y + q] = SCREENMEM[y + q] << 1;
                    SCREENMEM[y + q] |= SCREENMEM[y + 1 + q] >> 7;
                }
                SCREENMEM[3 + q] = SCREENMEM[3 + q] << 1;
                SCREENMEM[3 + q] |= OFFSCREEN[q] >> 7;
                for (y = 0; y < 3; y++) {
                    OFFSCREEN[y + q] = OFFSCREEN[y + q] << 1;
                    OFFSCREEN[y + q] |= OFFSCREEN[y + 1 + q] >> 7;
                }
                OFFSCREEN[3 + q] = OFFSCREEN[3 + q] << 1;
            }
        }
        wait_period(SCROLL_PERIOD);
    }
}

static void draw_image(unsigned char * xbm_bits, BYTE blank)
{
    /* XBM format. Must be 32x16 i.e. draw as portrait then 
     * rotate 90 clockwise before Saving */
    BYTE i, j, q;

    if (blank) draw_blank();

    q = SCR_SIZE - 1;
    for (i = 0; i < SCR_SIZE; i+=4) {
        for (j = 0; j < 4; j++) {
            OFFSCREEN[i+j] |= xbm_bits[q-j];
        }
        q -= 4;
    }
}

static void draw_well(void)
{
    BYTE i, j;
    BYTE x, y;
    char offsety = 0;
    BYTE tilt = is_tilted();

    // Lights out
    draw_blank();
    if (tilt) {
        draw_image(tilt_bits, 0);
    }

    // Well area
    for (y = 0; y < NUM_ROWS; y++) {
        for (x = 0; x < NUM_COLUMNS; x++) {
            if (get_well(y, x)) {
                set_pixel(x + 1, y);
            }
        }
        set_pixel(0, y);
        set_pixel(NUM_COLUMNS + 1, y);
    }
    for (x = 0; x < (NUM_COLUMNS + 2); x++) {
        set_pixel(x, NUM_ROWS);
    }

    // Incoming pieces
    for (i = 0; i < NUM_INCOMING; i++) {
        Tet_State * s = &incoming[i];
        char offsetx = 0;
        char low = 16;
        char high = 0;
        char left = 16;

        for (j = 0; j < SQUARE_NUM; j++) {
            char x = TETRITABLE[s->i][s->r][j][0];
            char y = TETRITABLE[s->i][s->r][j][1];
            if (y < low) low = y;
            if (x < left) left = x;
        }
        offsetx = SCR_WIDTH - 3;
        if ((s->i == I_PIECE) && (s->r == 0)) {
            // I-piece in horizontal orientation needs extra space
            offsetx --;
        }

        for (j = 0; j < SQUARE_NUM; j++) {
            char x = offsetx + TETRITABLE[s->i][s->r][j][0] - left;
            char y = offsety + TETRITABLE[s->i][s->r][j][1] - low;

            if (y > high) {
                high = y;
            }
            if ((y < (SCORE_START - 1)) && (x < SCR_WIDTH) && (x > 0)) {
                set_pixel(x, y);
            }
        }
        offsety = high + 2;
        if (offsety >= SCORE_START) break;
    }

    if (!tilt) {
        if (is_button_pressed()) {
            draw_number(0, SCORE_START, min_slack);
        } else {
            draw_number(1, SCORE_START, score);
        }
    }
}

static WORD get_period_for_level(void)
{
    WORD divisor = player_moves_per_down_move + 1;
    WORD level_50 = (level > 50) ? 50 : level;
    WORD dp2 = MAX_PERIOD - 
            (((MAX_PERIOD - MIN_PERIOD) * level_50) / 50);
    WORD tp2 = dp2 / divisor;

    if (tp2 < MIN_PERIOD) tp2 = MIN_PERIOD;
    if (tp2 > MAX_PERIOD) tp2 = MAX_PERIOD;
    return tp2;
}

#if 0
static void fill_up(void)
{
    BYTE budget = 0;

    // unfill the screen, pixel by pixel
    while(budget < 100) {
        BYTE x = drng() % SCR_WIDTH;
        BYTE y = drng() % SCR_HEIGHT;

        if (!get_pixel(x, y)) {
            budget++;
            continue;
        } else {
            budget = 0;
            clr_pixel(x, y);
        }
        flip();
        wait_period(10);
    }
    // flood fill
    draw_blank();
    flip();
}
#endif

void halt_and_catch_fire(BYTE info)
{
    BYTE i;

    draw_blank();
    draw_number(0, 0, info);
    draw_number(0, 8, score);
    draw_number(0, 16, min_slack);
    draw_number(0, 24, game_number);
    flip();
    while (1) {}
}

void Tetris(void)
{
    WORD hiscore = 0;
    WORD tilted_for = 0;
    Game_State rc = 0;
    WORD period = 0;
    BYTE i = 0;
    BYTE hand_of_god = 0;

    draw_blank();
    flip();

    // Hardware setup - for the tilt switch
    TRISEbits.TRISE0 = 1;   // E0 is input
    PORTEbits.REPU = 1;     // E0 has weak pullup
    TRISAbits.TRISA5 = 0;   // A5 is output
    LATAbits.LATA5 = 0;     // A5 is low

    // Software setup (global)
    drng_value = 0x31337;
    game_init(1);
    timing_error = 0;
    game_number = 0;

    while(1) {
        // "Hello" screen
        draw_image(hello_bits, 1);
        scroll_in();
        period = 0;
        do {
            if (is_button_pressed()) {
                Reset(); // go to bootloader
            }
            period += MIN_PERIOD;
            wait_period(MIN_PERIOD);
        } while (period < SCREEN_TIME);

        // Game select screen
        // Tilt to randomise the game number (initially 1)
        game_number++; 
        draw_image(game_bits, 1);
        draw_number(1, 16, game_number);
        scroll_in();

        period = 0;
        do {
            draw_image(game_bits, 1);
            if (is_tilted()) {
                game_number += drng();
                period = 0;
                draw_image(tilt_bits, 0);
            }
            if (is_button_pressed()) {
                game_number ++;
                period = 0;
            }
            game_number --;
            game_number %= 9999;
            game_number ++;
            draw_number(1, 16, game_number);
            flip();
            period += MIN_PERIOD;
            wait_period(is_button_pressed() ? MAX_PERIOD : MIN_PERIOD);
        } while (period < SCREEN_TIME);

        // Falling blocks screen 
        // Tilt to delete the current block and move into alternate speed mode
        draw_blank();
        game_init(game_number);
        game_reset();
        for (i = 0; i < 3; i++) {
            game_step();
        }
        draw_well();
        wait_period(0);
        min_slack = 999;
        hand_of_god = 0;
        scroll_in();

        do {
            rc = game_step();

            if (score > hiscore) {
                hiscore = score;
            }
            draw_well();

            period = get_period_for_level();

            if (is_button_pressed()) {
                level = 99;
                hand_of_god = 1;
            }
            if (is_tilted()) {
                if (tilted_for == 0) {
                    // initial action is delete
                    if (game_delete()) { 
                        for (i = 0; i < SCR_SIZE; i++) {
                            OFFSCREEN[i] ^= 0xff;   // flash screen
                        }
                        hand_of_god = 1;
                    }
                    period = FLASH_TIME;
                    tilted_for = period;
                } else if (tilted_for < HOLD_DOWN_TIME) {
                    period = MAX_PERIOD;   // slow down
                    tilted_for += period;   
                } else {
                    period = MIN_PERIOD;    // tilted for a while - speed up
                }
            } else {
                tilted_for = 0;
            }
            flip();
            wait_period(period);
        } while(rc != GAME_OVER);
    
        // End of game
        wait_period(SCREEN_TIME);

        // Game review screen
        draw_image(scores_bits, 1);
        draw_number(1, 9, score);
        draw_number(1, 23, hiscore);
        if (hand_of_god) {
            set_pixel(0, 0);
        }
        scroll_in();
        wait_period(SCREEN_TIME * 3);
    }
}


